// Automatic Tea Bag Steeper 

#define CUPTHRESH	((signed int16)thresh*10)

//motor direction states
#define STANDBY		0
#define LOWER		1
#define RAISE		2

//operational states
#define	INIT		0
#define BAGLOAD		1
#define	PLUNGE		2
#define	STEEP		3
#define	BREWED		4
#define GOSLEEP		5

#DEFINE _ON FALSE				//negative logic
#DEFINE _OFF TRUE				//negative logic

//Constants end.

#include <16F887.h>
//#device *=16			//Enable 16 bit pointers to yeild pointer access banks 2&3.
#device ADC=10			//get 10 bit results from read_adc()


#define XTAL 4000000
#use delay(clock=XTAL,RESTART_WDT)
#fuses XT,NOWDT,PUT,PROTECT,NOBROWNOUT,MCLR,NOLVP,NOCPD,NOWRT,NOFCMEN

//#use	rs232(baud=128000,bits=8,parity=N,xmit=PIN_C6,rcv=PIN_C7,ERRORS,BRGH1OK,stream=uart)

#use	fast_io(A)
#use	fast_io(B)
#use	fast_io(C)
#use 	fast_io(D)
#use 	fast_io(E)

//Declare special function register bytes
#byte	porta	=getenv("SFR:PORTA")
#byte	portb	=getenv("SFR:PORTB")
#byte	portc	=getenv("SFR:PORTC")
#byte	portd	=getenv("SFR:PORTD")
#byte	porte	=getenv("SFR:PORTE")

#byte	tmr1L	=getenv("SFR:TMR1L")
#byte	tmr1H	=getenv("SFR:TMR1H")
#byte	tmr2	=getenv("SFR:TMR2")
#byte	tmr3L	=getenv("SFR:TMR3L")
#byte	tmr3H	=getenv("SFR:TMR3H")
#byte	rcsta	=getenv("SFR:RCSTA1")
#byte	PRODL	=getenv("SFR:PRODL")
#byte	PRODH	=getenv("SFR:PRODH")
#byte	CCPTMRS1=getenv("SFR:CCPTMRS1")
#byte	APFCON	=getenv("SFR:APFCON")

#byte	CM2CON0	=getenv("SFR:CM2CON0")
#byte	CM2CON1	=getenv("SFR:CM2CON1")
#byte	VREFCON0=getenv("SFR:VREFCON0")
#byte	VREFCON1=getenv("SFR:VREFCON1")
#byte	VREFCON2=getenv("SFR:VREFCON2")

#byte	T1CON	=getenv("SFR:T1CON")
#byte	T1GCON	=getenv("SFR:T1GCON")
#byte	INTCON	=getenv("SFR:INTCON")

//Declare bits in SFR bytes
#bit	TMR2IF	=getenv("bit:TMR2IF")
#bit	TMR2IF	=getenv("bit:TMR2IF")
#bit	TMR3IF	=getenv("bit:TMR3IF")
#bit	carry	=getenv("bit:C")
#bit	T1GSEL	=getenv("bit:T1GSEL")
#bit	T1GGO	=getenv("bit:T1GGO")

//Declare bits in SFR bytes
//PORTA BITS
#bit	CUP_SNS	=porta.0		//cup sensor plate
#bit	ANA_1	=porta.1		//spare analogue
#bit	ANA_2	=porta.2		//spare analogue
#bit	MOTORB	=porta.3		//H-Bridge
#bit	MOTORA	=porta.4		//H-Bridge
#bit	LED		=porta.5		//Paddle LED anode

//PORTB BITS
#bit	PUSH_SW	=portb.0		//Paddle switch to ground
#bit	BOOST	=portb.1		//Boost enable for 6V6
#bit	LO_LIMIT=portb.2		//Tray low position to ground
#bit	HI_LIMIT=portb.3		//Tray high position to ground
#bit	SW_A	=portb.4		//quad sw
#bit	SW_B	=portb.5		//quad sw
#bit	SW_6	=portb.6		//
#bit	SW_7	=portb.7		//

//PORTC BITS (SERIAL I/O etc)
#bit	LEDA4	=portc.0		//Digit 1 Anodes
#bit	LEDA3	=portc.1		//Digit 2 Anodes
#bit	LEDA2	=portc.2		//Digit 3 Anodes
#bit	LEDA1	=portc.3		//Digit 4 Anodes
#bit	LEDKX	=portc.4		// :
#bit	LEDKT	=portc.5		// '
#bit	SER_TX	=portc.6		//Uart transmit
#bit	SER_RX	=portc.7		//Uart recieve
 
//PORTD BITS
#bit	LEDKA	=portd.0		//DISPLAY CATHODES
#bit	LEDKB	=portd.1		//
#bit	LEDKC	=portd.2		//
#bit	LEDKD	=portd.3		//
#bit	LEDKE	=portd.4		//
#bit	LEDKF	=portd.5		//
#bit	LEDKG	=portd.6		//
#bit	LEDKP	=portd.7		//

//PORTD BITS
#bit	WIFI	=porte.0		//Wifi enable
#bit	E1		=porte.1		//
#bit	E2		=porte.2		//

//Statics are automatically zeroed for us
static int	flags;
#bit	timesup	=flags.0
#bit	dial_cw	=flags.1
#bit	dial_ccw=flags.2
#bit	sync_rdy=flags.3
#bit	diagnostics=flags.4
#bit	powerup	=flags.5

//16 bit eeprom pointer required
#rom int16 getenv("EEPROM_ADDRESS") = {1,20}

static int		mode_timer;
static int		mode_is;
int				centisecs;
signed int16	cup_ref;
signed int16	clear_ref;
int16			sensor_raw;
float 			sensor_avg;

int 			itset;							//tensecs
int				thresh;							//cup sense threshold
static int 		old_sw_state;					//used to detect switch changes 
static int 		sync;							//display mux. sync down-counter providing 1Sec refresh flag

#INCLUDE "display.c"
#INCLUDE <math.h>

#inline
void	mux_display();
void	dialed(int);
void	teatime();

//1953.125 times/sec
#INT_TIMER0
void timer0_isr()
{
	mux_display();								//refresh multiplex display every 4ms	 (takes 4uS)
	if(--sync==0)								//50ms tick?
	{
		sync=97;//195;							//reload 50ms countdown
		sync_rdy=TRUE;							//flag to background
		centisecs++;							//rolling count

		//get time for last discharge and initiate another cup sensor detection cycle
		sensor_raw=get_timer1();				//read previous frozen discharge time latched in T1 (GATED)
		set_tris_A(0b00000000);					//set port A to output
		CUP_SNS=TRUE;							//apply voltage to charge-up the sensor plate
	}
	else if(sync==77)//92//190					//roughly 10ms later after plate charged...
	{
		set_timer1(5);							//begin timing again
		set_tris_A(0b00000001);					//set port A0 to input to comparator -ve and wait for comparator to freeze timer 1
	}
}

#INT_TIMER1
void timer1_isr()
{

}

#INT_RB
void switch_isr()
{
	int new_sw_state;

	new_sw_state=portb&0x30;		//mask out RB0 which is also enabled as INT_EXT!
	if(new_sw_state != old_sw_state) //Check for a change
	{
		if (bit_test(new_sw_state,5) == bit_test(old_sw_state,4))
 			dial_ccw=TRUE;
		else
			dial_cw=TRUE;
		old_sw_state=new_sw_state; // save state 
	}
}

#INT_EXT
void push_switch()
{

}

void main()
{
	porta=(0b00011000);		//initial data for port A (motor both HI)
	portb=(0b00000000);		//initial data for port B (dont care - all switch inputs)
	portc=(0b00110000);		//initial data for port C (clear all four anode driver bits and special segs (HARDWARE SPECIFIC))
	portd=(0b11111111);		//initial data for port D ()
	porte=(0b00000000);		//initial data for port E ()
	
//Set port I/O Direction (in/out = 1/0)
	set_tris_A(0b00000000);		//set port A to output (all driven low)
	set_tris_B(0b00000001);		//set port B I/O direction (all switches out except B0 INT input)
	set_tris_C(0b00000000);		//set port C I/O direction (RXD,SDA,FLOW in)
	set_tris_D(0b00000000);		//set port D to all o/p
	set_tris_E(0b00000000);		//set port E to O/P
	setup_adc_ports(NO_ANALOGS);
	port_B_pullups(0b00000001);	//just pull-up B0 INT to wake us from sleep

	powerup=TRUE;										//flag to display powerup info
	itset=read_eeprom(0);								//get stored time for last brew
	thresh=read_eeprom(1);								//get stored cup threshold for last brew
	setup_timer_0(T0_DIV_2);							//int 4MHz/(4*2*256) 1953.125 times/sec (~2KHZ)
	enable_interrupts(INT_EXT_H2L);						//let paddle switch wake us
	INTCON&=0b11111100;									//clear switch ints
	enable_interrupts(global);							//ready!

	do
	{
		sleep();											//xtal osc. & timer 0 shuts off
		//wake up...
		delay_cycles(1);									//NOP

		set_tris_B(0b00111101);								//set port B I/O direction (all switches in except B0 INT)
		port_B_pullups(0b00111101);							//just pull-up B0 INT to wake us from sleep
		old_sw_state=portb;									//match portb to clear interrupt
		INTCON&=0b11111100;									//clear switch ints
		enable_interrupts(INT_RB4|INT_RB5);					//let spin wheel adjust time

		//set up analog stuff
		setup_vref(VREF_HIGH | 15);							//VREF=3.95
		setup_comparator(CP2_A0_VREF | CP2_T1_SYNC);		//CP2_T1_GATE NOT!!! CCS GATE BIT INVERTED!!!
		setup_timer_1(T1_INTERNAL | T1_DIV_BY_1 | T1_GATE);	//div 1 to get readings ~1000
		setup_adc_ports(sAN0);								//analogue input to sensor plate

		//now let timer 0 drive display & sensing
		enable_interrupts(int_timer0);		
		teatime();
		disable_interrupts(int_timer0);

		//brew finished - go back into lowest possible power sleep mode
		portc=0b00110000;									//deactivate all four anode driver bits and special segs (HARDWARE SPECIFIC)
		portd=0b00000000;									//clear all segs
		CUP_SNS=FALSE;										//apply ground to sensor plate
		set_tris_A(0b00000000);								//set port A to output to save power
		set_tris_B(0b00000001);								//set port B I/O direction (all switches out except B0 INT input)
		setup_adc_ports(NO_ANALOGS);						//digital output
		port_B_pullups(0b00000001);							//pull-up B0 INT to wake us from sleep
		setup_vref(0);
		setup_comparator(NC_NC);
		setup_timer_1(T1_DISABLED);
	}
	while(TRUE);
}


int display_time(int16 secs)
{
	int16	mins;

	if(diagnostics)
	{
		printf(display,"%4lu\n",(int16)sensor_avg);
	}
	else
	{
		mins=secs/60;
		secs=secs%60;
		printf(display,"%2lu:%02lu\n",mins,secs);
	}
}

signed int16 sensor_reading()
{
	sensor_avg=(sensor_avg*0.9 + (float)sensor_raw*0.1);	//update avg
//fprintf(uart,"DATA,TIME,%f,%lu,%lu\r",sensor_avg,clear_ref,cup_ref);
	return (signed int16) sensor_avg;
}

//read-modify-write on porta was interfering with CUP_SNS so keep it high.
void bag_lift(int direction)
{
	switch(direction)
	{
		//BOOST only needed to raise gantry
		case RAISE:	 BOOST=TRUE;	porta=((porta&0b11100111)|0b00001001); break;	//MOTORA=0; MOTORB=1; CUP_SNS=1;
		case LOWER:	 BOOST=FALSE;	porta=((porta&0b11100111)|0b00010001); break;	//MOTORA=1; MOTORB=0; CUP_SNS=1;
		case STANDBY:BOOST=FALSE;	porta=((porta&0b11100111)|0b00011001); break;	//MOTORA=1; MOTORB=1; CUP_SNS=1;
	}
}

void paddle_led(short state)
{
	if(state==TRUE)
		porta=porta|0b00100001;		//read-modify-write on porta was interfering with CUP_SNS so keep it high.
	else
		porta=(porta&0b11011111)|1;	//read-modify-write on porta was interfering with CUP_SNS so keep it high.
}

//initialise avg and get cup absence reference
int initialise()
{
	int i;
	int	press;

	printf(display,"    \n");
	printf(display,"CHAI\a");
	diagnostics=FALSE;
	paddle_led(TRUE);										//light paddle
	bag_lift(RAISE);										//raise holder
	sync=1;													//preload to start conversion immediately
	press=0;												//prepare to count extra cycles added while paddle still pressed down
	for(i=0; i<40; i++)										//2 secs minimum
	{
		if(HI_LIMIT==_ON)	bag_lift(STANDBY);				//turn off motor when lift reaches the top limit
		sync_rdy=FALSE;										//clear flag for next 50ms tick
		while(!sync_rdy);									//wait for new tick

		if(i<3) sensor_avg=sensor_raw;						//initialise avg on second time round to get good timer value
		clear_ref=sensor_reading();							//find average initial sensor plate reading with no cups present

		if(PUSH_SW==_ON)									//paddle still pressed?
		{
			i=0;											//extend loop if we are still leaning on the switch
			if(++press==50)									//look for 2.5 second long press
			{
				diagnostics=TRUE;
				printf(display,"tESt\b");
				paddle_led(FALSE);							//turn off paddle LED to prompt paddle release
			}
		}
	}

	//normally reached top by now except if capacative sensor disconnected or faulty!
	while(HI_LIMIT==_OFF);									//ensure lift reaches the top limit
	bag_lift(STANDBY);										//ensure motor turned off

	//long "On" press engages diagnostic test mde
	if(diagnostics)
	{
		printf(display,"%4ld\t",clear_ref);
		delay_ms(2000);
	}

	return BAGLOAD;
}

//raise teabag holder, accept time adjustments, initiate plunge when paddle pressed
int load_teabags()
{
	int		display_hold;
	int		i;

	dial_cw=FALSE;					//ensure no twiddles buffered up
	dial_ccw=FALSE;					//
	display_hold=0;					//prepare to hold time display if dial twiddled


	//bag holders in position. Wait for plunge/time adjust
	do
	{
		if(dial_cw||dial_ccw)			//dial twiddled?
		{
			if(dial_cw)
			{
				dial_cw=FALSE;
				if(diagnostics)
				{
					thresh++;
					if(thresh>100) thresh=100;
				}	
				else
				{
					itset++;
					if(itset>60) itset=60;
				}
			}
			else
			{
				dial_ccw=FALSE;
				if(diagnostics)
				{
					thresh--;
					if(thresh<1) thresh=1;
				}
				else
				{
					itset--;
					if(itset<1) itset=1;
				}
			}
			display_hold=40;					//freeze display for 2000 centisecs
		}

		if(sync_rdy)							//new centisec?
		{
			sync_rdy=FALSE;						//clear flag for next centisec
			cup_ref=sensor_reading();			//get ref level for empty cups in position
			paddle_led((centisecs&8)==0);
			if((centisecs&8)||display_hold||diagnostics)		//4 centisec display time (blink paddle off) if not held
			{
				if(diagnostics&&display_hold)
				{
					printf(display,"%4ld\n",(int16)thresh*10);
				}
				else
				{
					display_time((int16)itset*10);	//convert 10's of seconds into secs for flashing mins:secs display
				}
				if(display_hold)	display_hold--;					//dial just twidlled?	countdown
			}
			else
			{
				printf(display,"  :  \n");		//blink out time
			}
		}
	}
	while(PUSH_SW==_OFF);						//loop until paddle pressed to plunge teabags

	if(diagnostics)
	{
		//store thresh if different to last time
		if(read_eeprom(1)!=thresh) write_eeprom(1,thresh);	
	}
	else
	{
		//store time if different to last time
		if(read_eeprom(0)!=itset) write_eeprom(0,itset);
	}
	return PLUNGE;								//go into plunging mode
}

int plunge_bags()
{
	int i;

	paddle_led(TRUE);						//paddle lit while bags plunge
	display_time((int16)itset*10);			//convert 10's of seconds into secs for steady mins:secs display
	bag_lift(LOWER);						//lower the teabag holder
	while(LO_LIMIT==_OFF);					//wait until holder reaches the bottom limit
	bag_lift(STANDBY);						//turn off motor

	//prepare for long press abort
	i=0;
	while(PUSH_SW==_ON)
	{
		delay_ms(100);
		if(++i>10) 
		{
			paddle_led(FALSE);				//paddle goes dark
			printf(display,"    \n");
			while(PUSH_SW==_ON);
			return GOSLEEP;
		}	
	}
	return STEEP;							//go into steeping mode
}

int steep_bags()
{
	signed int16 presence;
	int16 timer;
	int stir;
	int	cs;

	//detect water being poured to start timer
	paddle_led(FALSE);							//paddle LED off during steeping
	do
	{
		sync_rdy=FALSE;							//clear flag for next centisec
		while(!sync_rdy);						//wait for new centisec
		presence=sensor_reading()-cup_ref;		//test for presence of cup filled with water

		if(((centisecs&31)<22)||diagnostics)					//16 centisec display time
		{
			display_time((int16)itset*10);		//convert 10's of seconds into secs for flashing mins:secs display
		}
		else
		{
			printf(display,"FILL\n");			//blink "fill"
		}
	}
	while((PUSH_SW==_OFF)&&(presence<CUPTHRESH));

	//water poured (or switch pressed to abort)
	if(PUSH_SW==_OFF)							//not aborted?
	{
		paddle_led(diagnostics);				//only lit if in diagnostics
		timer=itset;							//start timer
		timer*=10;								//time in seconds
		cs=20;									//0.1sec countdown
		stir=60;
		do
		{
			sync_rdy=FALSE;						//clear flag for next centisec
			while(!sync_rdy);					//wait for new centisec
			presence=sensor_reading();			//keep up the averaging
			if((--cs==0)||diagnostics)
			{

				timer--;						//decrement by one second
				display_time(timer);			//update mins:secs display
				if(timer==3) WIFI=TRUE;			//give advanced warning
				if(--stir==0)					//after 1 min
				{
					stir=60;
					bag_lift(RAISE);
					delay_ms(250);
					bag_lift(STANDBY);			//turn off motor
					delay_ms(750);
					bag_lift(LOWER);
					while(LO_LIMIT==_OFF);		//wait until holder reaches the bottom limit
					bag_lift(STANDBY);			//turn off motor
					cs=1;						//
				}
				else
					cs=20;						//reload cs for 1sec countdown
			}
		}
		while((timer>0)&&(PUSH_SW==_OFF));	

		return BREWED;							//return to bag unload mode to lift bags from teacups
	}
	else
	{
		return BAGLOAD;							//return to bag load mode to lift bags from teacups and redo if aborted
	}
}

//steeping terminated
int finished()
{
	signed int16 presence;
	int16 timeout;
	//WIFI=TRUE;									//enable ESP8266
	paddle_led(TRUE);							//paddle lit while teabag holder rising
	bag_lift(RAISE);							//raise holder
	while(HI_LIMIT==_OFF);						//wait until lift reaches the top limit
	bag_lift(STANDBY);							//turn off motor

	delay_ms(1000);								//pause to let mechs. settle before reading cup sensor
	timeout=2400;								//centisec display timeout 2min
	do
	{
		sync_rdy=FALSE;							//clear flag for next centisec
		while(!sync_rdy);						//wait for new 50ms
		presence=sensor_reading()-clear_ref;	//test for presence of cup filled with water (many rapid calls to span 50Hz)

		if(timeout>0)
		{
			if(--timeout==0)
				printf(display,"    \a");		//2mins passed... blank out display
			else
			{
				if(((centisecs&31)<22)||diagnostics)	//alternate between time display
				{
					display_time((int16)itset*10);		//convert 10's of seconds into secs for flashing mins:secs display
					paddle_led(FALSE);
				}
				else
				{
					printf(display,"CHAI\n");			//blinking CHAI
					paddle_led(TRUE);
				}
			}
		}
		else									//more thn 2min and nobody arrived to remove cups
		{
			paddle_led((centisecs&31)==0);		//just blink paddle
		}
	}
	while((PUSH_SW==_OFF)&&(presence>CUPTHRESH));	//remain in raised position until switch operated or cups removed

	paddle_led(TRUE);
	printf(display,"    \n");		//display off
	if(PUSH_SW==_OFF)				//if auto presence det...
		delay_ms(3000);				//add a few seconds to allow cup to be clear of teabag holder
	bag_lift(LOWER);				//lower the teabag holder
	while(LO_LIMIT==_OFF);			//wait until holder reaches the bottom limit
	bag_lift(STANDBY);				//turn off motor
	paddle_led(FALSE);				//paddle goes dark
	WIFI=FALSE;						//disable ESP8266
	return GOSLEEP;
}

//processor woken by paddle push switch
void teatime()
{
	if(powerup)
	{
		powerup=FALSE;
		printf(display,"%4lu\n",(int16)thresh*10);
		delay_ms(1000);
	}
	mode_is=INIT;	//prepare for cup placement and teabag loading
	do
	{
		centisecs=0;	//resync timer for display animations
		switch(mode_is)
		{
			case INIT:		mode_is=initialise();	break;
			case BAGLOAD:	mode_is=load_teabags();	break;
			case PLUNGE:	mode_is=plunge_bags();	break;
			case STEEP:		mode_is=steep_bags();	break;
			case BREWED:	mode_is=finished();		break;
		}
	}
	while(mode_is!=GOSLEEP);
}